/******************************************************************************

	Image Acquisition & Display Program - Primitive Control Functions

	Exposes two cameras simultaneously, then displays the 
	images serially after waiting for a keypress between images.
	Also allows the images to be written to a file.
	
					Bernard Nemeth - June 10, 1992
					(based on work by Kev Forrester)
					
******************************************************************************/
#include <dos.h>
#include <bios.h>
#include <conio.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <io.h>
#include <sys\types.h>
#include <sys\stat.h>

#include <at200.h>
#include <camera.h>

#include "idecls.h"
#define  MAIN_PROG			/* For camsys.h */
#include "camsys.h"

/* Function Prototypes. */

void mcga_display_mode( int );
void mcga_color_set( int, int, int, int );
void select_camera ( int );
void load_exposure_ops( void );
void load_transfer_ops( void );
void show_image( void );
void show_help( void );
void main( int, char *[] );
void init_disp ( void );
void init_struct ( void );

/* Defines */

#define DEFAULT_CCD       4		/* Thomson 512 x 512 CCD */
#define DEFAULT_ANALOG    32		/* 200KHz 12 bit analog electronics */
#define DEFAULT_EXPOSURE  800 		/* Exposure time in milliseconds */
#define DEFAULT_SHUTTER   40		/* Shutter open/close time in ms. */
#define DEFAULT_DATA_BITS 12		/* Number of data bits/pixel */
#define DEFAULT_BINNING	  1		/* 1=Bin 1x1, else Bin MAX x MAX */
#define DEFAULT_GAIN	  0		/* 0 = 1X gain,  1 = 4X gain */

/* Set default parameters. */

int ccd_type    = DEFAULT_CCD,
    ser_byte    = DEFAULT_ANALOG,
    data_bits   = DEFAULT_DATA_BITS,
    shutter     = DEFAULT_SHUTTER,
    p_bin       = DEFAULT_BINNING,
    s_bin       = DEFAULT_BINNING,
    ems_handle  = 0,
    disp_sw     = 0,
    gain        = DEFAULT_GAIN;
    
unsigned long exposure   = DEFAULT_EXPOSURE;
    
char	f1name[32];	/* Output file name for camera 1. */
char	f2name[32];	/* Output file name for camera 2. */
char	fname[32];

/* screen_row[] is initialized in main() to point to the start of each row on
 *  the VGA display.  It is then used by the show_image() function to update
 *  the VGA memory. 
 */

char far *screen_row[200] = { 0, };   /* Pointers to graphics display rows */

int far *ems_buff;			/* EMS data page */

/* Structure definition of data file header. 
 *
 * Contains data to identify details of a given exposure.
 */

   /* File header (fixed size 512 bytes) for file. */ 
 
   union file_hdr_union 
   {
      struct cam_structs_tag 
      { 	
         cam_misc misc;
         cam_sys  sys;
         cam_exp  exp;
      } structs;
      char     filler[512];
   } file_hdr;
   

/* Select text mode or 320 x 200 - 256 color graphics mode */

void mcga_display_mode( int mode )
{ 
union REGS regs;
  regs.x.ax = (mode) ? 19 : 3;  /* AH=0, AL=Mode */
  int86(16, &regs, &regs);      /* Call video bios */
}

/* Set the RGB values for 64 level grey scale display */

void mcga_color_set( int color, int red, int green, int blue )
{ 
  outp(0x3C8,color);  /* DAC color control register selects the color */
  outp(0x3C9,red);    /* Set Red DAC (0..63) */
  outp(0x3C9,green);  /* Set Green DAC (0..63) */
  outp(0x3C9,blue);   /* Set Blue DAC (0..63) */
}

/* Select a camera by setting all camera-specific addresses. */

void select_camera ( int camera_number )
{
     sel_cam = camera_number;
     if ( camera_number == 0 )
     {
    	sel_sys  = &std0_sys;	/* Select camera 0 */
    	sel_exp  = &std0_exp;
    	sel_misc = &std0_misc;
        at200_base = 0x220;
     }
     else if (camera_number == 1 )
     {
    	sel_sys  = &std1_sys;	/* Select camera 1 */
    	sel_exp  = &std1_exp;
    	sel_misc = &std1_misc;
        at200_base = 0x230;
     }
}

/* Setup DSP operations to take an exposure. */

void load_exposure_ops( void )
{
unsigned x;

  disable_trigger();
  outp(at200_base + SER_CONTROL,ser_byte);
  x = at200_base + CONTROL;
  outp(x,inp(x) & ~FIFO_INPUT_SELECT);
  dsp_command( SET_PARALLEL_DELAY );
  dsp_arg( (long) ccd[ccd_type].parallel_delay );
  dsp_command( SET_PARALLEL_STATES );
  dsp_arg( (long) ccd[ccd_type].parallel_states );
  dsp_command( SET_SHUTTER_CLOSE );
  dsp_arg( (long) shutter );
  dsp_command( SET_SHUTTER_OPEN );
  dsp_arg( (long) shutter );
  dsp_command( LOAD_OPERATIONS );
  dsp_op( OP_SET_GAIN, gain );
  dsp_op( OP_EXPOSE_SCALE, (unsigned int) exposure ); /* Exposure Timer Scale in msec */
  dsp_op( OP_PARALLEL_DIS,ccd[ccd_type].rows );
    dsp_op( OP_OPEN_SHUTTER,0 );
    dsp_op( OP_EXPOSE,1 );            /* Expose 1 * Timer Scale */
    dsp_op( OP_CLOSE_SHUTTER,0 );
  dsp_op( OP_TERMINATE,0 );
 
  /* Wait for DSP to complete initialization */
 
  for( x = 1; 
       x && (inp(at200_base+DSP_INTERRUPT_STATUS)&(HOST_FLAG_2|HOST_FLAG_3));
       x++ )
  ;    /* NULL statement */
  
  dsp_command( EXECUTE );	/* Execute the above script. */
}

/* Setup DSP operations to start data transfer from camera. */

void load_transfer_ops( void )
{
unsigned x, rows, cols;

  rows = ccd[ccd_type].rows / p_bin;
  cols = ccd[ccd_type].cols / s_bin;
  
  dsp_command( LOAD_OPERATIONS );
    x = ccd[ccd_type].cols + ccd[ccd_type].prescan + ccd[ccd_type].postscan;
    dsp_op( OP_SERIAL_DIS, x * 5 );     /* Clear Serial Register */
    dsp_op( OP_SET_SER_BIN, s_bin );
    dsp_op( OP_DO_LOOP, rows );
      dsp_op( OP_PARALLEL_BIN, p_bin );
      dsp_op( OP_SERIAL_DIS, ccd[ccd_type].prescan );
      dsp_op( OP_SERIAL_CONV, cols );
      dsp_op( OP_SERIAL_DIS, ccd[ccd_type].postscan );
    dsp_op( OP_END_DO,0 );
  dsp_op( OP_TERMINATE,0 );
  
  /* Wait for DSP to complete initialization */
  
  for( x = 1; 
       x && (inp(at200_base+DSP_INTERRUPT_STATUS)&(HOST_FLAG_2|HOST_FLAG_3));
       x++ )
  ; 
  
  _dos_getdate ( &sel_misc->dt );	/* Date and time of exposure. */
  _dos_gettime ( &sel_misc->tm );
  dsp_command( EXECUTE );	/* Do it! */
}

void show_image( void )
{
unsigned char buff[320], row;
int      max, min, ifact;
int      page, u, s;
register int  x, y;
register long z;
long          t;
int	      fd, count;

  reset_fifo();		/* Clear FIFO for start of input. */
  load_transfer_ops();	/* Start data transfer. */
  
  /* Wait for 1st Pixel */ 

  t = 0xFFFFFL;		/* Timeout count */
  while( (!(inp( at200_base + STATUS) & FIFO_EMPTY) && ( --t )) )  
  ;
  if ( !t )		/* Timed out. */
  {
    printf( "Timeout waiting for data detected in show_image for controller %X.\n", 
            at200_base );
    x = getchar();			/* wait for user to press any key. */
    return;
  }
  
  /* Get data from camera, hold in EMS memory */
  
  ems_collect_image( ems_handle );

  /* Check for overflow */
  
  if( !(inp(at200_base + STATUS) & FIFO_OVERFLOW) ) 
  {
    printf( "FIFO overflow detected in show_image for controller %X.\n", 
            at200_base );
    exit( 1 );
  }
  
  if ( disp_sw )	/* Display image? */
  { 
     /* Set up VGA to 320 X 200 graphics mode. */

     mcga_display_mode( 1 );
     for( x=0 ; x<64 ; x++ )
        mcga_color_set( x, x, x, x );


     /* Scan image, get min and max pixel intensities. */
     
     t = 262144L / ( p_bin * s_bin );	/* # of binned pixels to display */
     max = -32768;
     min = 32767; 
  
     for( y=8192, page=0, z=0; z < t; y++, z++ ) 
     {
        if( y == 8192 ) 		/* At EMS page boundary. */
        {
           ems_map_page( 0, ems_handle, page++ );
     	y = 0;
        }
        if( ems_buff[y] > max )
           max = ems_buff[y];
        if( ems_buff[y] < min )
           min = ems_buff[y];
     }
     
     sel_misc->pixel_min = min;
     sel_misc->pixel_max = max;
     
     /* Calculate intensity expansion factor. 
      * Result is such that screen color values result in a range 0..63 
      * selecting 64 grey scale levels.
      */

     ifact = (( max - min + 62 ) / 63 );
     if( !ifact )
	ifact = 1;

     /* Copy image from EMS memory to display, adjust pixel range. */
   
     if ( t > 64000L )
        t = 64000L;
     u = s = 512 / s_bin;
     if ( s > 320 )
        s = 320;

     for( y=8192, row=0, x=0, page=0, z=0L; z < t; x++, y++, z++ ) 
     {
     	if( y >= 8192 ) 
     	{
     		ems_map_page( 0, ems_handle, page++ );
     		y -= 8192;
     	}
     	
        buff[x] = (unsigned char) (( ems_buff[y] - min) / ifact );
        if( x >= ( s - 1 ) ) 
        {
           memcpy( screen_row[row++], buff, s );
           if ( row == 200 )
              break;			/* Displayed 200 rows. */
           x = -1;
           if ( u > 320 )		/* If at rhs of video buff */
              y += ( u - 320 );		/* discard rhs bits of camera image */
        }
     }
     x = getchar();			/* wait for user to press any key. */
     mcga_display_mode( 0 );		/* then reset VGA from graphics mode. */
  }  					/* if (disp_sw ) */
  
  if ( fname[0] )	/* Write file of image data? */ 
  {
     fd = open( fname, O_WRONLY | O_CREAT |O_BINARY , S_IREAD | S_IWRITE );
     
     /* Create file header data, and write it out. */
     
     memset ( &file_hdr, ' ', 512 );	/* Clear file header */

     /* Copy camera structures to file header, then write it out. */

     memcpy ( &file_hdr.structs.misc, sel_misc, sizeof( cam_misc ) );
     memcpy ( &file_hdr.structs.sys,  sel_sys,  sizeof( cam_sys ) );
     memcpy ( &file_hdr.structs.exp,  sel_exp,  sizeof( cam_exp ) );

     count = write ( fd, &file_hdr, 512 );
     if ( count != 512 )
     {
     	printf ( "Unable to write file header (disk full ?)\n" );
     	exit ( -1 );
     }
   
     for( page=0; page < ( 32 / ( s_bin * p_bin ) ); page ) 
     {
        ems_map_page( 0, ems_handle, page++ );
        count = write( fd, ems_buff, 16384 );    /* write one EMS memory page */
        if ( count != 16384 )
        {
     	   printf ( "Unable to write file data (disk full ?)\n" );
           exit ( -1 );
        }
     }
     close( fd );
  }
}


void show_help(void)
{

  printf("USAGE: display [-aN #] [-bN ] [-eN #] [-o #] [-g #] [-fN xxx]\n"
         "  Where: N = camera number ( 1 or 2 ) \n\n"
         "    -aN #  # is the base address descriptor (2 by default)\n"
         "    -bN #  # binning factor. (2, 3, 4 - default 1)\n"
         "    -eN #  # is the exposure time in milliseconds (>0)\n"
         "    -o  #  # is the open/close time for the shutter in ms. (>0)\n"
         "    -g  #  # sets preamp gain.  0 = 1X gain,  1 = 4X gain.\n"
         "    -fN xxx  write data for camera N to file xxx\n" 
         "    -d       display the images.\n\n" );
  exit( 0 );
}

void main( int argc, char *argv[] )
{
register unsigned x;

  init_disp();

  f1name[0] = '\0';			/* File names for cameras */
  f2name[0] = '\0';

  /******************************************************************/
  
  /* Check for command line arguments. */

  for( x=1 ; x < (unsigned) argc ; x++ ) {
    if( argv[x][0] == '-' )
      switch( argv[x][1] ) {
        case 'b': case 'B':
          if ( argv[x][2] == '1' )
             std0_exp.sbin = std0_exp.pbin = atoi( argv[ ++x ] );
          else if ( argv[x][2] == '2' )
              std1_exp.sbin = std1_exp.pbin = atoi( argv[ ++x ] );
          else
             show_help();
          break;
        case 'd': case 'D':
          disp_sw = 1;
          break;
        case 'g': case 'G':
          gain = atoi( argv[ ++x ] );
          break;
        case 'e': case 'E':
          if ( argv[x][2] == '1' )
             std0_exp.exposure = atoi( argv[ ++x ] );
          else if ( argv[x][2] == '2' )
             std1_exp.exposure = atoi( argv[ ++x ] );
          else
             show_help();
          if( ( std0_exp.exposure <= 0 ) || ( std1_exp.exposure <= 0 ) )
            show_help();
          break;
        case 'f': case 'F':
          if ( argv[x][2] == '1' )
            strcpy( f1name, argv[ ++x ] );
          else if ( argv[x][2] == '2' )
            strcpy( f2name, argv[ ++x ] );
          else
             show_help();
          break;
        case 'o': case 'O':
          x++;
          shutter = atoi(argv[x]);
          if( shutter <= 0 )
            show_help();
          break;
        default:
          show_help();
          break;
      }
    else
      show_help();
  }

  /******************************************************************/

  /* Select camera 0 */
  
  select_camera ( 0 );
  p_bin = sel_exp->pbin;
  s_bin = sel_exp->sbin;
  exposure   = sel_exp->exposure;

  ems_handle = ems_alloc ( 262144 * sizeof( unsigned ) );
  if ( !ems_handle )
  {
  	printf( "Unable to allocate EMS memory - check system configuration.\n" );
  	exit( -1 );
  }
  
  ems_buff = (unsigned *) ems_get_page_frame();
  if ( !ems_buff )
  {
  	printf( "Unable to locate EMS page frame - check system configuration.\n" );
  	exit( -1 );
  }
  
  for( x=0 ; x<200 ; x++ )    /* Load graphics display array */
    screen_row[x] = &((char far *)0xA0000000)[x * 320];

  load_exposure_ops();

  /******************************************************************/
  
  /* Select camera 1 */
  
  select_camera ( 1 );
  p_bin = sel_exp->pbin;
  s_bin = sel_exp->sbin;
  exposure   = sel_exp->exposure;

  if ( f2name[0] )
    strcpy( fname, f2name );
  else
    fname[0] = '\0';
  
  load_exposure_ops();

  /******************************************************************/
  
  /* Wait for controller to reach idle state. */

  while(!( inp( at200_base + DSP_INTERRUPT_STATUS) & (HOST_FLAG_2|HOST_FLAG_3)))
  ;

  while(   inp( at200_base + DSP_INTERRUPT_STATUS) & (HOST_FLAG_2|HOST_FLAG_3))
  ;

  /******************************************************************/
  
  /* Display image of camera 1 */

  show_image();

  /******************************************************************/
  
  /* Display image of camera 0 */

  select_camera ( 0 );
  p_bin = sel_exp->pbin;
  s_bin = sel_exp->sbin;

  if ( f1name[0] )
    strcpy( fname, f1name );
  else
    fname[0] = '\0';
  

  show_image();

  /******************************************************************/

  /* All done - terminate. */
  
  ems_free( ems_handle );
  exit( 0 );
}

/******************************************************************/

void init_disp ( void )	/* Initialize camera structures. */
{
  char *env;
  
  /* Check for EMM capability. */
  
  if ( !emm_present() )
  {
  	printf( "EMM capability not present - check system configuration.\n" );
  	exit( -1 );
  } 
  
  if ( emm_version() < 0x40 )
  {
  	printf( "EMM version incorrect - check system configuration.\n" );
  	exit( -1 );
  } 
  
  ems_buff = (unsigned *) ems_get_page_frame();
  if ( !ems_buff )
  {
  	printf( "Unable to locate EMS page frame - check system configuration.\n" );
  	exit( -1 );
  }
  
  /* Initialize camera 1 structures. */
 
  select_camera( 1 );
  init_struct();

  /* then camera 0 */

  select_camera( 0 );
  init_struct();

  /* Get environment variables for path names. */

  if ( ( env = getenv( "KI_IMAGE" ) ) != (char *) NULL )
  {
     strcpy( imagepath, env );
     strcat( imagepath, "\\" );
  }
  else
     imagepath[0] = '\0';

}

void init_struct ( void )
{
  unsigned int x;
  
  sel_sys->ccd = ccd[4];	/* CCD type - TH7985B 512 X 512 pixels */
  if ( sel_cam == 0 )
  {
     sel_sys->at200_id = 2;	/* Controller ID and base addr */
     at200_base = 0x220;
  }
  else
  {
     sel_sys->at200_id = 3;
     at200_base = 0x230;
  }
  sel_sys->frame = 0;		/* This is a full frame CCD */
  sel_sys->open_delay = 5;	
  sel_sys->close_delay = 5;
  sel_sys->preamp_delay = 7;
  sel_sys->speed = 2;		/* 200K Hz Analog to digital convertor */
  sel_sys->data_bits = 14;	/* with 14 bit/pixel resolution */
  
  sel_exp->exposure = 200;	/* default exposure */
  sel_exp->gain = 0;		/* 0 = 1X, 1 = 4X gain */
  sel_exp->mpp_mode = 1;	/* allow/disallow MPP exposures */
  sel_exp->corg = 0;		/* sub-array col origin, row origin */
  sel_exp->rorg = 0;
  sel_exp->cols = 512;		/* subarray # of rows, cols */
  sel_exp->rows = 512;
  sel_exp->pbin = 2;		/* Column binning factor */
  sel_exp->sbin = 2;		/* Row binning factor */
  sel_exp->clr_cnt = 5;		/* # of times to clear CCD before exposure */
  sel_exp->row_shift = 0;

  sel_misc->camera = (unsigned char) sel_cam;	/* camera number */
  sel_misc->preamp = 0;		/* Preamp on/off during exposure. */
  sel_misc->ems_handle = 0;	/* EMS handle for aquired data */
  sel_misc->exp_scale = 1;	/* Exposure scale 1=ms, 1000 = sec, 60000=min */
  for ( x=0; x<5; x++ )	 	/* Temperature readings from AFB */
  {
     sel_misc->temp.tei[x] = 0;
     sel_misc->temp.tee[x] = 0;
     sel_misc->temp.td[x]  = 0;
     sel_misc->temp.tcc[x] = 0;
     sel_misc->temp.tcf[x] = 0;
     sel_misc->temp.tfw[x] = 0;
  }
  sel_misc->sequence = 0;		/* file sequence # */
  sel_misc->exposure_type = 'A';	/* Airglow/Aurora, Bias, Dark, Flat */
  sel_misc->version = (1 * 16) + 0;	/* file/program version = 1.0 */

  dsp_command( DSP_HALT );		/* Stop any operation in progress. */

  /* Load setup program into camera */
    
  x = load_dsp_program( dsp_cam200k, size_of_dsp_cam200k );

  if( !x ) 
  {
    printf("ERROR: Unable to load program into AT200 at addr %X\n", at200_base );
    exit( 1 );
  }

  /* Wait for DSP to complete initialization */
  
  for( x=1; x; x++ )
    if( !(inp( at200_base+DSP_INTERRUPT_STATUS) & (HOST_FLAG_2 | HOST_FLAG_3)) )
      break;  
  if( !x ) 
  {  
    printf("AT200 at addr %X is unresponsive or missing.  Unable to proceed.\n",
           at200_base );
    exit( 2 );
  }
}


